from zope.i18n import translate
from Products.validation import validation
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ZopeTwoPageTemplateFile

from collective.sendaspdf.emailer import send_message

from collective.sendaspdf import SendAsPDFMessageFactory as _
from collective.sendaspdf.browser.base import BaseView


class SendForm(BaseView):
    """ If the user does not have Javascript enabled, he is
    redirected to this page.
    This page generates the PDF using the URL of the previous page
    and shows a form to send it by mail.

    It should logically show the same thing than the previous page
    (the user has no JS enabled so no funky stuff happened).
    The only case it could fail is if the previous page was processing a
    POST form. In this case, we can just render the page as if it was
    called with a simple GET request.
    """
    error_mapping = {'email': ['email', 'invalid_email'],
                     'email_recipient': ['email_recipient',
                                         'invalid_email_recipient'],
                     'pdf_name': ['file_not_found',
                                  'file_not_specified',
                                  'file_unauthorized']}

    def check_form(self):
        """ Checks the form submitted when the user clicks on
        the 'send by mail' button.
        """
        form = self.request.form
        user = self.get_user()
        self.check_pdf_accessibility()

        fields = ['name_recipient',
                  'email_recipient',
                  'title',
                  'text']
        if not user:
            fields.extend(['name', 'email'])

        # All fields are mandatory
        for field in fields:
            if not form.get(field):
                self.errors.append(field)

        # We check that emails are real emails.
        email_validator = validation.validatorFor('isEmail')
        email_fields = ['email_recipient']
        if not user:
            email_fields.append('email')

        for field in email_fields:
            value = form.get(field)
            if not value:
                continue
            if not email_validator(value) == 1:
                self.errors.append('invalid_' + field)

    def get_values(self):
        """ Provides the values used to fill the form.
        """
        if self.errors:
            return self.request.form

        values = {'pdf_name': self.filename}

        values['title'] = self.pdf_tool.mail_title
        values['text'] = self.pdf_tool.mail_content

        if self.get_user():
            values['name'] = self.get_user_fullname()
            values['email'] = self.get_user_email()

        return values

    def process_form(self):
        """
        """
        form = self.request.form

        if self.get_user():
            mfrom = '%s <%s>' % (self.get_user_fullname(),
                                 self.get_user_email())
        else:
            mfrom = '%s <%s>' % (form['name'],
                                 form['email'])

        mto = '%s <%s>' % (form['name_recipient'],
                           form['email_recipient'])
        pdf_file = file('%s/%s' % (self.tempdir,
                                   form['pdf_name']),
                        'r')
        success = send_message(
            mfrom,
            mto,
            form['title'],
            form['text'],
            pdf_file,
            self.pdf_tool.filename_in_mail)
        return success

    def get_editor(self):
        """ Get's the editor to use in the send form.
        """
        memberdata = getToolByName(self.context,
                                   'portal_memberdata')
        default_editor = memberdata.get('wysiwyg_editor', None)

        if not default_editor:
            # In Plone4, default editor is in
            # portal_properties/site_properties/default_editor.
            portal_props = getToolByName(self.context,
                                         'portal_properties')
            try:
                default_editor = portal_props.site_properties.default_editor
            except AttributeError:
                default_editor = 'plone_wysiwyg'

        portal_membership = getToolByName(self.context,
                                          'portal_membership')
        member = portal_membership.getAuthenticatedMember()
        if member:
            member_editor = member.getProperty('wysiwyg_editor',
                                               default_editor)
        editor = member_editor or default_editor
        return editor.lower()

    def __call__(self):
        form = self.request.form

        if 'form_submitted' in form:
            # The user clicked on the 'send by mail'
            # button.
            self.check_form()
            if not self.errors:
                success = self.process_form()
                if success:
                    msg = _(u'msg_success',
                            default=u'The e-mail has been sent')
                    msg_type = 'info'
                else:
                    msg = _(u'msg_send_fail',
                            default=u'Could not send the e-mail')
                    msg_type = 'error'
            else:
                msg = _(u'msg_error',
                        default=u'Errors appeared while processing your form')
                msg_type = 'error'

            self.context.plone_utils.addPortalMessage(
                translate(msg,
                          target_language=self.get_lang()),
                type=msg_type)

            if not self.errors:
                self.request.response.redirect(self.context.absolute_url())

        elif 'form_cancelled' in form:
            # The user clicked on the 'cancel' button.
            self.request.response.redirect(self.context.absolute_url())
        else:
            # The user clicked on the 'send by mail'
            # link.
            self.make_pdf()
            if self.errors:
                # The PDF generation did not work, we render the page
                # used when errors are found when sending the mail.
                self.index = ZopeTwoPageTemplateFile('templates/download.pt')

        # We need the self parameter for Plone 4
        return self.index(self)



import os
from datetime import datetime

from zope.i18n import translate
from AccessControl import ClassSecurityInfo
from persistent.dict import PersistentDict
from zope.annotation.interfaces import IAnnotations
from zope.interface import Interface
from zope.interface import implements
from Products.Archetypes import atapi
from Products.CMFCore.utils import ImmutableId
from Products.CMFCore.permissions import ModifyPortalContent
from Products.ATContentTypes.content.document import ATDocument
from Products.ATContentTypes.content.document import ATDocumentSchema

import config

from collective.sendaspdf import SendAsPDFMessageFactory as _


class ISendAsPDFTool(Interface):
    """send as pdf tool marker interface"""


sendAsPDFSchema = ATDocumentSchema.copy() + atapi.Schema((
    atapi.StringField(
        name='pdf_generator',
        default='wk',
        widget=atapi.SelectionWidget(
            format="select",
            label=_(u'label_pdf_generator',
                    default=u'PDF generator')),
        vocabulary='_generatorVocabulary'),

    atapi.StringField(
        name='tempdir',
        default='/tmp',
        widget=atapi.StringWidget(
            label=_(u'label_export_dir',
                    default=u'Directory where PDF files will be stored'))),

    atapi.LinesField(
        name='excluded_browser_attachment',
        default=[],
        widget=atapi.LinesWidget(
            label=_(u'label_excluded_browsersattachments',
                    default=u'List of browsers for which PDF '
                    'filename will not be forced'),
            description=_(u'label_help_excluded_attachments',
                          default=u'Browsers might warn the user '
                          'that downloading the PDF might harm their '
                          'computer. You can set here a list of browsers '
                          'for which the system will not force the download. '
                          'We recommend to have Chrome in the list'))),

    atapi.BooleanField(
        name='auto_cleanup_pdf',
        default=True,
        widget=atapi.BooleanWidget(
            label=_(u'label_auto_cleanup_pdf',
                    default=u'Auto cleanup'),
            description=_(u'help_auto_cleanup_pdf',
                          default=u'Automatic cleanup of temporarily '
                          'generated pdf files'))),

    atapi.BooleanField(
        name='rewrite_https_links',
        default=False,
        widget=atapi.BooleanWidget(
            label=_(u'label_rewrite_https_links',
                    default=u'Rewrite https links'),
            description=_(
                u'help_rewrite_https_links',
                default=u'Replace https links by http links. '
                        u'If style sheets or images are not visible '
                        u'in the pdf, this option may help. '
                        u'This is a simple replacement of "https://" '
                        u'in the entire page source, so we may replace '
                        u'more than you want.'))),

    atapi.StringField(
        name='salt',
        default='salt_as_pdf',
        widget=atapi.StringWidget(
            label=_(u'label_salt',
                    default=u'SALT used when hashing e-mails of users'))),

    atapi.StringField(
        name='mail_title',
        widget=atapi.StringWidget(
            label=_(u'label_mail_title',
                    default=u'Default title of e-mails')),
        schemata='mail'),

    atapi.TextField(
        name='mail_content',
        default_content_type='text/html',
        allowable_content_types=('text/html', ),
        widget=atapi.RichWidget(
            label=_(u'label_mail_content',
                    default=u'Default body of e-mails')),
        schemata='mail'),

    atapi.StringField(
        name='filename_in_mail',
        default='screenshot.pdf',
        widget=atapi.StringWidget(
            label=_(u'label_filename_title',
                    default=u'Name of the PDF file in the mail')),
        schemata='mail'),

    atapi.BooleanField(
        name='always_print_css',
        widget=atapi.BooleanWidget(
            label=_(u'label_print_css_always',
                    default=u'Always use print CSS'),
            description=_(u'help_print_css_always',
                          default=u'Always use the print CSS (only valid '
                          'with wkhtmltopdf, xhtml2pdf will use the '
                          'print CSS whatever you chose)')),
        schemata='wk'),

    atapi.BooleanField(
        name='allow_cookie',
        widget=atapi.BooleanWidget(
            label=_(u'label_allow_cookie',
                    default=u'Allow passing cookies to sendaspdf'),
            description=_(u'help_allow cookie',
                          default=u'Allow passing the user\'s '
                          'cookie to wkhtmltopdf')),
        schemata='wk'),

    atapi.LinesField(
        name='print_css_types',
        widget=atapi.LinesWidget(
            label=_(u'label_print_css',
                    default=u'Portal types using print css'),
            description=_(u'help_print_css',
                          default=u'You can register here a list of portal '
                          'types for which the system must use the print '
                          'CSS instead of the screen one (currently only '
                          'working if you use wkhtmltopdf. xhtml2pdf uses '
                          'the print CSS by default). One type per line')),
        schemata='wk'),

    atapi.BooleanField(
        name='use_book_style',
        widget=atapi.BooleanWidget(
            label=_(u'label_use_book_style',
                    default=u'Use book style'),
            description=_(u'help_use_book_style',
                          default=u'PDF generated will have a book '
                          'style (will override custom margins). '
                          'This is not supported by all wkhtmltopdf '
                          'versions.')),
        schemata='wk'),

    atapi.BooleanField(
        name='generate_toc',
        widget=atapi.BooleanWidget(
            label=_(u'label_add_toc',
                    default=u'Add table of contents'),
            description=_(u'help_add_toc',
                          default=u'A table of contents will be '
                          'prepended to the generated files.')),
        schemata='wk'),

    atapi.IntegerField(
        name='margin_top',
        default=10,
        widget=atapi.IntegerWidget(
            label=_(u'label_margin_top',
                    default=u'Margin top')),
        schemata='wk'),

    atapi.IntegerField(
        name='margin_right',
        default=10,
        widget=atapi.IntegerWidget(
            label=_(u'label_margin_right',
                    default=u'Margin right')),
        schemata='wk'),

    atapi.IntegerField(
        name='margin_bottom',
        default=10,
        widget=atapi.IntegerWidget(
            label=_(u'label_margin_bottom',
                    default=u'Margin bottom')),
        schemata='wk'),

    atapi.IntegerField(
        name='margin_left',
        default=10,
        widget=atapi.IntegerWidget(
            label=_(u'label_margin_left',
                    default=u'Margin left')),
        schemata='wk'),
))

# Hides the default fields.
for field in ['title', 'description', 'text']:
    if field in sendAsPDFSchema:
        sendAsPDFSchema[field].widget.visible = {'edit': 'invisible',
                                                 'view': 'invisible'}

# Hides the fields other than the ones we defined.
for key in sendAsPDFSchema.keys():
    if sendAsPDFSchema[key].schemata not in ['default', 'mail', 'wk']:
        sendAsPDFSchema[key].widget.visible = {'edit': 'invisible',
                                               'view': 'invisible'}


class SendAsPDFTool(ImmutableId, ATDocument):
    """ Tool for sendaspdf product.
    Allows to set various settings:
    - the PDF generator used.
    - the directory to store PDF files.
    - the SALT used in md5 hashing
    - the default title for the mail
    - the default body of the mail
    """
    security = ClassSecurityInfo()
    __implements__ = ()
    implements(ISendAsPDFTool)

    id = 'portal_sendaspdf'
    typeDescription = "Configure send as pdf"
    typeDescMsgId = 'description_edit_sendaspdftool'
    schema = sendAsPDFSchema

    def __init__(self, *args, **kwargs):
        self.setTitle('Send as PDF configuration')

    security.declareProtected(ModifyPortalContent, 'indexObject')

    def indexObject(self):
        pass

    security.declareProtected(ModifyPortalContent, 'reindexObject')

    def reindexObject(self, idxs=[]):
        pass

    security.declareProtected(ModifyPortalContent, 'reindexObjectSecurity')

    def reindexObjectSecurity(self, skip_self=False):
        pass

    security.declarePublic('_genderVocabulary')

    def _generatorVocabulary(self):
        return atapi.DisplayList([
            ('pisa',
             _(u'label_pisa',
               default=u'XHTML2PDF: HTML/CSS to PDF converter written '
               'in Python (http://www.xhtml2pdf.com/)')),
            ('wk',
             _(u'label_wk',
               default=u'wkhtmltopdf: Simple shell utility to convert '
               'html to pdf using the webkit (Safari, Chrome) rendering '
               'engine (http://code.google.com/p/wkhtmltopdf/)'))])

    def _getMetadata(self):
        """ Gets the annotations linked to the tool.
        """
        anno_key = 'collective.sendaspdf'
        annotations = IAnnotations(self)

        metadata = annotations.get(anno_key,
                                   None)
        if metadata is None:
            annotations[anno_key] = PersistentDict()
            metadata = annotations[anno_key]

        return metadata

    def getPDFList(self):
        """ Gets the list of PDF generated by the product.
        """
        metadata = self._getMetadata()
        if 'pdf_files' not in metadata:
            metadata['pdf_files'] = PersistentDict()

        return metadata['pdf_files']

    def setPDFList(self, new_list):
        metadata = self._getMetadata()
        metadata['pdf_files'] = new_list

    def get_last_clean(self):
        metadata = self._getMetadata()
        if 'last_clean' not in metadata:
            metadata['last_clean'] = datetime.now()

        return metadata['last_clean']

    def registerPDF(self, filename):
        """ Registers a PDF filename in the list if auto_cleanup is enabled.
        """
        auto_cleanup = self.getAuto_cleanup_pdf()
        if not auto_cleanup:
            return

        now = datetime.now()
        pdfs = self.getPDFList()

        pdfs[filename] = now

        if (now - self.get_last_clean()).seconds > 1000:
            self.cleanPDFs()

    def cleanPDFs(self):
        """ Removes from the filesystem the list of PDF
        files generated more than 2 hours ago.
        """
        to_delete = []
        pdfs = self.getPDFList()
        now = datetime.now()

        # First we compute the list of files to delete.
        for filename in pdfs:
            filedate = pdfs[filename]
            delta = now - filedate

            if delta.seconds > 7200:
                to_delete.append(filename)

        existing_files = os.listdir(self.tempdir)
        for filename in to_delete:
            del pdfs[filename]
            if filename in existing_files:
                os.remove('%s/%s' % (self.tempdir,
                                     filename))
        self.setPDFList(pdfs)
        metadata = self._getMetadata()
        metadata['last_clean'] = now

    def is_browser_excluded(self, browser_name):
        """ Returns True if the browser id (got from
        the user agent) falls under a rule in
        'excluded_browser_attachment'
        """
        for excluded in self.getExcluded_browser_attachment():
            if excluded in browser_name:
                return True

        return False

    def make_options(self):
        """ Returns a dictionnary of options that can be used in
        the tool.
        """
        if self.pdf_generator == 'pisa':
            return {}

        toc_msg = _(u'label_toc',
                    default=u'Table of content')

        options = {
            'book': self.getUse_book_style(),
            'margin-top': self.getMargin_top(),
            'margin-right': self.getMargin_right(),
            'margin-bottom': self.getMargin_bottom(),
            'margin-left': self.getMargin_left()
        }

        gen_toc = self.getGenerate_toc()
        if gen_toc:
            options['toc'] = gen_toc
            options['toc-header-text'] = translate(
                toc_msg, context=self.REQUEST)

        ac_cookie = self.REQUEST.cookies.get('__ac', None)
        if self.getAllow_cookie() and ac_cookie is not None:
            options['cookie'] = ['__ac', ac_cookie]

        return options

atapi.registerType(SendAsPDFTool, config.PROJECTNAME)



# Defines the html_to_pdf method using PISA:
# http://github.com/holtwick/xhtml2pdf

simple_options = []
valued_options = []


def html_to_pdf(source, export_dir, filename, original_url,
                use_print_css, extra_options=[]):
    # We import pisa inside the function so it does not raise
    # import exception if pisa is not installed.
    try:
        # First try newer version of library under a different name.
        import xhtml2pdf.pisa as pisa
        pisa  # pyflakes
    except ImportError:
        # Try the old library.  Note that you can also get the above
        # ImportError simply because you are on Python 2.4 and
        # xhtml2pdf.pisa tries and fails to import hashlib.
        import ho.pisa as pisa

    file_path = '%s/%s' % (export_dir, filename)

    pdf_file = file(file_path, "wb")
    link_callback = pisa.pisaLinkLoader(original_url).getFileName

    pdf = pisa.CreatePDF(source,
                         pdf_file,
                         log_warn=1,
                         log_err=1,
                         path=original_url,
                         link_callback=link_callback,
                         )

    if pdf.err:
        return None, pdf.err

    return pdf_file, None



from zope import component
from zope.interface import implements
from OFS.interfaces import IFolder

from collective.sendaspdf.interfaces import ISendAsPDFOptionsMaker


class FolderOptionsMaker(object):
    implements(ISendAsPDFOptionsMaker)
    component.adapts(IFolder)

    def __init__(self, context):
        self.context = context

    def overrideAll(self):
        return False

    def getOptions(self):
        return {'header-center': 'This was added by a custom adapter',
                'margin-top': 42,
                'footer-center': self.context.title,
                'book': 1,
                '--no-toc': 1}



from zope.interface import Interface


class ISendAsPDFOptionsMaker(Interface):
    """
    """

    def overrideAll():
        """ Returns a boolean telling if the options
        specified by the adapter should override the settings
        provided by the user or specified in the request.
        """

    def getOptions(context=None):
        """ Returns a dictionary of options that will be applied
        to generated PDF.
        """



from collective.sendaspdf.browser.base import BaseView


class PreDownloadPDF(BaseView):
    """ This page is the one called when clicking on the
    'Download as PDF' view.
    It generates the PDF file then redirects to the real
    download view (see below).
    """
    def __call__(self):
        self.make_pdf()
        if self.errors:
            return self.index(self)

        self.request.form['pdf_name'] = self.filename
        return self.context.restrictedTraverse('@@send_as_pdf_download')()


class DownloadPDF(BaseView):
    """ View called when clicking the 'Click here to preview'
    link.
    """
    def generate_pdf_name(self):
        """ Generates the name for the PDF file.
        If the context title does not contain non-ascii characters,
        we'll use it.
        Otherwise we'll rewrite it using normalize string.
        """
        try:
            name = self.context.title.encode('ascii')
        except (UnicodeDecodeError, UnicodeEncodeError, ):
            name = self.context.id

        return '%s.pdf' % name

    def __call__(self):
        form = self.request.form
        self.check_pdf_accessibility()

        if self.errors:
            return self.index(self)

        self.pdf_file = file('%s/%s' % (self.tempdir,
                                        form['pdf_name']),
                             'r')
        self.request.response.setHeader("Content-type",
                                        "application/pdf")
        self.request.response.setHeader("X-Robots-Tag",
                                        "noindex")
        self.request.response.setHeader("Cache-Control",
                                        "no-cache, must-revalidate")

        user_agent = self.request['HTTP_USER_AGENT']
        if not self.pdf_tool.is_browser_excluded(user_agent):
            disposition = ('attachment; filename="%s"' %
                           self.generate_pdf_name())
            self.request.response.setHeader('Content-Disposition', disposition)

        return self.pdf_file.read()



# This file is taken from Products.PasswordResetTool
# We copied it here to avoid test problems if the product
# is not installed in the buildout.
# The product can be found here: http://plone.org/products/passwordresettool

from datetime import datetime, timedelta

from Products.Five.browser import BrowserView
from Products.SecureMailHost.SecureMailHost import SecureMailHost as MailBase


class LongView(BrowserView):
    """ A view that takes a loooooong time
    to answer (well, about 20 seconds).
    Used to ensure we correctly kill wkhtmltopdf
    when it takes too long to finish its job (that may lead
    to lock the Zope thread and getting wkhtmltopdf
    unusable until the next restart of the instance).
    """

    def wait(self):
        end_time = datetime.now() + timedelta(0.0002)
        while datetime.now() < end_time:
            pass

        return 'Ho, finished :)'


class MockMailHost(MailBase):
    """A MailHost that collects messages instead of sending them.

    Thanks to Rocky Burt for inspiration.
    """

    def __init__(self, id):
        MailBase.__init__(self, id, smtp_notls=True)
        self.reset()

    def reset(self):
        self.messages = []

    def send(self, message, mto=None, mfrom=None, subject=None, encode=None):
        """
        Basically construct an email.Message from the given params to make sure
        everything is ok and store the results in the messages instance var.
        """
        self.messages.append(message)

    def secureSend(self, message, mto, mfrom, **kwargs):
        kwargs['debug'] = True
        result = MailBase.secureSend(self, message=message, mto=mto,
                                     mfrom=mfrom, **kwargs)
        self.messages.append(result)

    def validateSingleEmailAddress(self, address):
        return True



# Code taken from PDFMiner tool pdf2txt:
# https://github.com/euske/pdfminer/blob/master/tools/pdf2txt.py

from pdfminer.pdfinterp import PDFResourceManager
from pdfminer.pdfinterp import PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams

from StringIO import StringIO


def display_pdf(content):
    fp = StringIO(content)
    codec = 'utf-8'
    laparams = LAParams()
    rsrcmgr = PDFResourceManager()

    outfp = StringIO()
    device = TextConverter(
        rsrcmgr, outfp, codec=codec, laparams=laparams)
    interpreter = PDFPageInterpreter(rsrcmgr, device)

    try:
        pagenos = set()
        for page in PDFPage.get_pages(
                fp,
                pagenos,
                maxpages=0,
                password='',
                caching=True,
                check_extractable=True):
            interpreter.process_page(page)
        fp.close()
    except:
        print 'Unable to process the PDF file'

    device.close()

    print '------------------'
    outfp.seek(0)
    print outfp.read()

    outfp.close()



from HTMLParser import HTMLParser


class SendAsPDFHtmlParser(HTMLParser):
    """ Extracts the data generated by collective.sendaspdf
    in a readable way.
    """

    def __init__(self, *args, **kwargs):
        HTMLParser.__init__(self, *args, **kwargs)

        # List of nodes encountered
        self.path = []

        # useful data stored while parsing the page.
        self.document_actions = []

        # A few flags: they should all be reseted to False at the end.
        self.in_dl_as_pdf_action = False
        self.in_send_as_pdf_action = False

    def get_document_actions(self):
        """ Returns self.document_actions but casts everything
        to str to avoid errors in tests due to a difference
        like u'blabla' instead of 'blabla'.
        """
        return [dict([(str(k), str(v)) for k, v in action.items()])
                for action in self.document_actions]

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)

        tag_id = attrs.get('id', '')
        tag_classes = attrs.get('class', '')
        self.path.append({'tag': tag,
                          'id': tag_id,
                          'class': tag_classes})

        if tag == 'li':
            if tag_id == 'document-action-download_as_pdf':
                self.in_dl_as_pdf_action = True

            if tag_id == 'document-action-send_as_pdf':
                self.in_send_as_pdf_action = True

        if tag == 'a':
            if self.in_dl_as_pdf_action or self.in_send_as_pdf_action:
                self.document_actions.append({})
                self.document_actions[-1]['href'] = attrs.get('href', '')
                self.document_actions[-1]['text'] = ''

    def handle_endtag(self, tag):
        if not self.path:
            return

        t = self.path.pop()
        tag_id = t.get('id', '')

        if t['tag'] == 'li':
            if tag_id == 'document-action-download_as_pdf':
                self.in_dl_as_pdf_action = False

            if tag_id == 'document-action-send_as_pdf':
                self.in_send_as_pdf_action = False

    def handle_startendtag(self, tag, attrs):
        attrs = dict(attrs)

    def handle_data(self, data):
        if not self.path:
            return

        tag = self.path[-1]

        if tag['tag'] == 'a':
            if self.in_dl_as_pdf_action or self.in_send_as_pdf_action:
                self.document_actions[-1]['text'] += data



# -*- coding: utf-8 -*-
import doctest

from Testing import ZopeTestCase as ztc

try:
    from Zope2.App import zcml
    is_plone_3 = False

    zcml  # pyflakes
except ImportError:
    # Plone 3
    from Products.Five import zcml
    is_plone_3 = True

try:
    from Testing.testbrowser import Browser

    Browser  # pyflakes
except ImportError:
    # Plone 3
    from Products.Five.testbrowser import Browser

import transaction
from Products.Five import fiveconfigure
from Products.PloneTestCase import PloneTestCase as ptc

ptc.setupPloneSite()

from zope.component import getSiteManager
from Acquisition import aq_base
from Products.MailHost.interfaces import IMailHost


import collective.sendaspdf
from collective.sendaspdf.tests.sendaspdf_parser import SendAsPDFHtmlParser
from collective.sendaspdf.tests.utils import MockMailHost


OPTIONFLAGS = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)


class SendAsPDFTestCase(ptc.FunctionalTestCase):
    is_plone_3 = is_plone_3

    # We'll use those pages taken from Wikipedia pages about Plone
    # to create some pages in the system.
    # We chose English for its lack of accented characters, French
    # because it mixes normal and accented characters and Japanese
    # for the use of non-Latin characters.
    # (my apologies to Japanese people if I make the text sound weird
    # by cutting it at 80 characters, I don't really have a clue about
    # what the text means).
    data = {'en': """
            <p>Plone is a free and open source content management system
            built on top of the Zope application server. In principle,
            Plone can be used for any kind of website, including blogs,
            internet sites, webshops and internal websites. It is also well
            positioned to be used as a document publishing system and groupware
            collaboration tool. The strengths of Plone are its flexible and
            adaptable workflow, very good security, extensibility, high
            usability and flexibility.</p>
            <p>Plone is released under the GNU General Public License (GPL) and
            is designed to be extensible. Major development is conducted
            periodically during special meetings called Plone Sprints.
            Additional functionality is added to Plone with Products,
            which may be distributed through the Plone website or otherwise.
            The Plone Foundation holds and enforces all copyrights and
            trademarks.
            Plone also has legal backing from the council of the Software
            Freedom Law Center.</p>
            <p>MediaWiki's "Monobook" layout is based partially on the
            Plone style sheets. High-profile public sector users include
            the Federal Bureau of Investigation, Brazilian Government, United
            Nations, City of Bern (Switzerland), New South Wales Government
            (Australia), and European Environment Agency.""",

            'fr': """
            <p>Plone est un système de gestion de contenu Web libre publié
            selon les termes de la GNU GPL. Il est construit au-dessus du
            serveur d'applications Zope et de son extension CMF
            (Content Management Framework).</p>
            <p>Plone est entièrement objet et modulaire. Il permet de
            créer et gérer les aspects d'un site web, comme les utilisateurs,
            les « workflows » ou les documents attachables.</p>
            <p>Le fait que Plone soit programmé sur Zope fait qu'il est
            moins utilisé que d'autres systèmes de gestion de contenu,
            Zope étant connu pour sa courbe d'apprentissage assez lente
            (le fameux ZopeZen). Toutefois, Plone est considéré comme étant
            l'un des meilleurs CMS open-source existants, et est utilisé entre
            autres par le site web de la FSF, la Fondation pour le logiciel
            libre.</p>
            """,

            'jp': """
            <p>PloneはZopeアプリケーションサーバ上に構築されたフリーかつオープンなコ
            ンテンツマネジメントシステムである。基本的にPloneは、ブログ、インターネットサイト
            、ウェブショップや組織内のWebサイトといったどのような用途にも利用できる。さらに文
            書管理システムやグループウェアといった共有ツールとしても利用できるようになっている
            。Ploneの強みは、その柔軟で適応性のあるワークフロー、非常に優れたセキュリティ、拡張性
            、高いユーザビリティと柔軟性である</p>
            <p>PloneはGNU General Public License(GPL)ライセンスでリリースされており
            、拡張し易いように設計されている。主な開発は、定期的に開催される Plone スプリン
            トと呼ばれる特別なミーティング(ハッカソン)で行われる。Ploneに対する追加機能はプ
            ロダクトで追加する。それはPloneのウェブサイトかその他のサイトで配布されている。
            Plone財団は全ての著作権と商標を保持している。さらにPloneはSoftware Freedom
            Law Centerの評議会からの法的な後ろ盾も持っている。</p>
            <p>MediaWikiの"Monobook [JP]"レイアウトは部分的にPloneのスタイルシートをベー
            スにしている。</p>
            """}

    def afterSetUp(self):
        self.portal._original_MailHost = self.portal.MailHost
        self.portal.MailHost = mailhost = MockMailHost('MailHost')
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(mailhost, provided=IMailHost)

        # his hack allows us to get the traceback when the getting a
        # 500 error when using the browser.
        self.portal.error_log._ignored_exceptions = ()

        def raising(self, info):
            import traceback
            traceback.print_tb(info[2])
            print info[1]

        from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
        SiteErrorLog.raising = raising
        transaction.commit()

    def beforeTearDown(self):
        self.portal.MailHost = self.portal._original_MailHost
        sm = getSiteManager(context=self.portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(aq_base(self.portal._original_MailHost),
                           provided=IMailHost)

    def login_as_user(self, username, password):
        self.browser.open('http://nohost/plone/logout')
        self.browser.open('http://nohost/plone/login_form')
        self.browser.getControl(name='__ac_name').value = username
        self.browser.getControl(name='__ac_password').value = password
        self.browser.getControl(name='submit').click()

    def login_as_manager(self):
        self.login_as_user(
            ptc.portal_owner,
            ptc.default_password)

    def install_products(self):
        self.browser = Browser()

        fiveconfigure.debug_mode = True
        zcml.load_config('configure.zcml',
                         collective.sendaspdf)
        zcml.load_config('tests.zcml',
                         collective.sendaspdf)

        ztc.installPackage(collective.sendaspdf)
        self.addProfile('collective.sendaspdf:default')
        self.addProfile('collective.sendaspdf:tests')
        self.addProduct('collective.sendaspdf')

    def create_folder(self, parent, title):
        self.login_as_manager()
        self.browser.open('%s/createObject?type_name=Folder' %
                          parent.absolute_url())
        self.browser.getControl(name='title').value = title
        self.browser.getControl(name='form.button.save').click()
        return self.browser.url

    def create_page(self, language, title=None, parent=None):
        self.login_as_manager()
        if parent is None:
            parent = self.portal

        self.browser.open('%s/createObject?type_name=Document' %
                          parent.absolute_url())
        if title is None:
            title = 'Plone (%s)' % language

        self.browser.getControl(name='title').value = title
        self.browser.getControl(name='text').value = self.data.get(language,
                                                                   'Oups ...')
        self.browser.getControl(name='form.button.save').click()
        return self.browser.url

    def create_image(self, img_file, img_file_type, parent=None):
        self.login_as_manager()
        if parent is None:
            parent = self.portal

        def mydir():
            import sys
            import os.path

            if __name__ == '__main__':
                filename = sys.argv[0]
            else:
                filename = __file__
            return os.path.abspath(os.path.dirname(filename))

        self.browser.open('%s/createObject?type_name=Image' %
                          parent.absolute_url())
        file_control = self.browser.getControl(name='image_file')
        file_control.mech_control.add_file(file('%s/%s' % (mydir(), img_file)),
                                           img_file_type,
                                           img_file)
        self.browser.getControl(name='form.button.save').click()
        return self.browser.url

    def setup_data(self):
        self.install_products()
        for lang in self.data.keys():
            self.create_page(lang)

        # We also add a page for which the title will contain
        # non-ASCII characters.
        # Depending on Plone version, the generated ID is not the same,
        # so we first generate the page with a simple title for which id
        # is safe, then we'll edit the title.
        self.cyrilic_title_page_url = self.create_page(None,
                                                       'Cyrilic title page')
        self.browser.open('%s/edit' % self.cyrilic_title_page_url)
        self.browser.getControl(name='title').value = '3цдса'
        self.browser.getControl(name='form.button.save').click()

    def get_sendaspdf_actions(self):
        """ Finds the links generated by send as pdf in browser contents.
        """
        parser = SendAsPDFHtmlParser()
        parser.feed(self.browser.contents)
        return parser.get_document_actions()

    def list_available_controls(self, form_name, before=None, hidden=[]):
        """
        'before' can be used to print something before the form.
        That can be useful for example when the first line printed
        starts with a non-deterministic value as you cannot
        use an ellipsis.
        """

        if before is not None:
            print before

        # Code from the community manual (page not available anymore,
        # sorry for the source :/ )
        form = self.browser.getForm(name=form_name)
        # get and print all controls
        for ctrl in form.mech_form.controls:
            try:
                control = self.browser.getControl(name=ctrl.name)
                if control.name in hidden:
                    continue

                print "%s: %s (%s)" % (control.name,
                                       control.value,
                                       control.type)
                print control.controls
            except:
                pass



##############################################################################
# Stollen from watcher list
# https://github.com/collective/collective.watcherlist
##############################################################################

import logging
from email.MIMEText import MIMEText
from email.MIMEBase import MIMEBase
from email.MIMEMultipart import MIMEMultipart
from email import Encoders
from smtplib import SMTPException
import pkg_resources
import socket

from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import getSiteEncoding
from Products.CMFPlone.utils import safe_unicode
try:
    from zope.component.hooks import getSite
    getSite  # pyflakes
except ImportError:
    # BBB for Plone 3
    from zope.app.component.hooks import getSite
from zope.component import getMultiAdapter

zope2_egg = pkg_resources.working_set.find(
    pkg_resources.Requirement.parse('Zope2'))
USE_SECURE_SEND = True
if zope2_egg and (zope2_egg.parsed_version >=
                  pkg_resources.parse_version('2.12.3')):
    USE_SECURE_SEND = False

logger = logging.getLogger('collective.sendaspdf')
DEFAULT_CHARSET = 'utf-8'


def get_charset():
    """Character set to use for encoding the email.

    If encoding fails we will try some other encodings.  We hope
    to get utf-8 here always actually.

    The getSiteEncoding call also works when portal is None, falling
    back to utf-8.  But that is only on Plone 4, not Plone 3.  So we
    handle that ourselves.
    """
    charset = None
    portal = getSite()
    if portal is None:
        return DEFAULT_CHARSET
    charset = portal.getProperty('email_charset', '')
    if not charset:
        charset = getSiteEncoding(portal)
    return charset


def su(value):
    """Return safe unicode version of value.
    """
    return safe_unicode(value, encoding=get_charset())


def get_mail_host():
    """Get the MailHost object.

    Return None in case of problems.
    """
    portal = getSite()
    if portal is None:
        return None
    request = portal.REQUEST
    ctrlOverview = getMultiAdapter((portal, request),
                                   name='overview-controlpanel')
    mail_settings_correct = not ctrlOverview.mailhost_warning()
    if mail_settings_correct:
        mail_host = getToolByName(portal, 'MailHost', None)
        return mail_host


def prepare_mail_message(msg, attachment, filename):
    """ Creates the message
    """
    html = msg

    # First, we transform the message in a text format.
    portal_transforms = getToolByName(getSite(),
                                      'portal_transforms')
    plain = portal_transforms.convert('html_to_text',
                                      html).getData()

    # We definitely want unicode at this point.
    plain = su(plain)
    html = su(html)

    # We must choose the body charset manually.  Note that the
    # goal and effect of this loop is to determine the
    # body_charset.
    for body_charset in 'US-ASCII', get_charset(), 'UTF-8':
        try:
            plain.encode(body_charset)
            html.encode(body_charset)
        except UnicodeEncodeError:
            pass
        else:
            break

    # Encoding should work now; let's replace errors just in case.
    plain = plain.encode(body_charset, 'replace')
    html = html.encode(body_charset, 'xmlcharrefreplace')

    text_part = MIMEText(plain, 'plain', body_charset)
    html_part = MIMEText(html, 'html', body_charset)

    # As we have plain text, html and attachment, we need to do
    # two multiparts:
    # - the first one contains the message
    # the second one includes the previous one and the attachment.
    email_content = MIMEMultipart('alternative')
    email_content.epilogue = ''
    email_content.attach(text_part)
    email_content.attach(html_part)

    # Now the attachment.
    attach = MIMEBase('application', 'pdf')
    attach.set_payload(attachment.read())
    Encoders.encode_base64(attach)
    attach.add_header('Content-Disposition',
                      'attachment; filename="%s"' % filename)

    # We attach everything to the mail.
    email_msg = MIMEMultipart()
    email_msg.attach(email_content)
    email_msg.attach(attach)
    return email_msg


def send_message(mfrom, mto, subject, message, attachment, filename):
    """Send an email.

    Return True when all is fine, False on problems.
    """
    message = prepare_mail_message(message, attachment, filename)

    mail_host = get_mail_host()
    if mail_host is None:
        logger.warn('Could not send email: mail host not defined.')
        return False

    mfrom = su(mfrom)
    mto = su(mto)
    subject = su(subject)
    header_charset = get_charset()

    try:
        if USE_SECURE_SEND:
            mail_host.secureSend(message=message,
                                 mto=mto,
                                 mfrom=mfrom,
                                 subject=subject,
                                 charset=header_charset)
        else:
            mail_host.send(message,
                           mto=mto,
                           mfrom=mfrom,
                           subject=subject,
                           immediate=True,
                           charset=header_charset)

    except (socket.error, SMTPException):
        logger.warn('Could not send email to %s with subject %s',
                    mto, subject)
        return False
    except:
        raise
    # sent successfully
    return True



# This package defines transformers for html to pdf.
# Each module defines a html_to_pdf method that takes three parameters:
# - source: the HTML source
# - export_dir: path to the directory where the PDF will be created.
# - filename: the name of the file (it the name does not end with '.pdf',
#   the method will add it.
# - original_url: the URL of the file being transformed.
# - use_print_css: tell if the system should use the print CSS or the screen
#   one (only supported by wk so far)
#
# Methods returns a <file> object and a list of potential errors.

import pisa
import wk

# pyflakes
pisa
wk



import os
from datetime import datetime
from random import randint

from Products.Five.browser import BrowserView
from Products.CMFCore.utils import getToolByName

from collective.sendaspdf import transforms
from collective.sendaspdf.utils import find_filename, update_relative_url

from collective.sendaspdf.utils import md5_hash
from collective.sendaspdf.utils import extract_from_url
from collective.sendaspdf.interfaces import ISendAsPDFOptionsMaker


class BaseView(BrowserView):
    """ Class used to factorize some code for the different views
    used in the product.
    """
    error_mapping = {}

    def __init__(self, *args, **kwargs):
        """ We just need to define some instance variables.
        """
        super(BaseView, self).__init__(*args, **kwargs)

        # The list of errors found when checking the form.
        self.errors = []

        # We get the configuration from the portal_sendaspdf
        self.pdf_tool = getToolByName(self.context,
                                      'portal_sendaspdf')
        self.tempdir = self.pdf_tool.tempdir
        self.salt = self.pdf_tool.salt
        self.pdf_generator = self.pdf_tool.pdf_generator
        self.filename_in_mail = self.pdf_tool.filename_in_mail

        self.pdf_file = None
        self.filename = ''

    def get_lang(self):
        """ Finds the language to use.
        """
        props = getToolByName(self.context,
                              'portal_properties')
        return props.site_properties.getProperty('default_language') or 'en'

    def get_user(self):
        """ Returns the currently logged-in user.
        """
        mtool = getToolByName(self.context, 'portal_membership')
        if mtool.isAnonymousUser():
            return

        return mtool.getAuthenticatedMember()

    def get_user_fullname(self):
        """ Returns the currently logged-in user's fullname.
        """
        member = self.get_user()
        if member:
            return member.getProperty('fullname')

    def get_user_email(self):
        """ Returns the currently logged-in user's email.
        """
        member = self.get_user()
        if member:
            return member.getProperty('email')

    def generate_filename_prefix(self):
        """ Returns the user's email hashed in md5 followed
        by an underscore.
        If we can not get an email (the user is anonymous or
        email is not mandatory in the system), returns an empty
        string.

        We extract it from 'generate_temp_filename as we will
        also use this sytem to be sure that the user has
        access to the file when sending it.
        """
        email = self.get_user_email()
        if not email:
            return ''

        return '%s_' % md5_hash(email, self.salt)

    def get_page_source(self):
        """ Returns the HTML source of a web page, considering
        that the URL of the page is contained in the form under
        the 'page_url' key.
        """

        if 'page_url' not in self.request.form:
            self.errors.append('no_source')
            return

        url = self.request.form['page_url']
        context_url = self.context.absolute_url()

        view_name, get_params = extract_from_url(url, context_url)

        # Now we will reinject the GET parameters in the request.
        if get_params:
            for key in get_params:
                self.request.form[key] = get_params[key]

        try:
            view = self.context.restrictedTraverse(view_name)
        except:
            # For some reason, 'view' as view name can fail (at least
            # in a client project in Plone 3.3.5), where calling the
            # same url in a browser works just fine...  The error here is:
            # AttributeError: 'view' object has no attribute '__of__'
            # Just take the context as view then.

            # Note that this has the same effect as calling
            # self.context.restrictedTraverse(x), where x is None or
            # an empty string.
            view = self.context

        return update_relative_url(view(), self.context)

    def generate_temp_filename(self):
        """ Generates the filename used to store the PDF file.
        Basically the md5 hash of th user's email followed
        by a timestamp.
        If the user is anonymous, just use the timestamp.
        In case of conflict, we just happen '-x' at the end.
        """
        prefix = self.generate_filename_prefix()
        now = datetime.now()
        # Ok that might not be the best timestamp system, but it's
        # enough for our needs.
        timestamp = '-'.join([
            ''.join([str(x) for x in now.timetuple()]),
            str(now.microsecond),
            str(randint(10000, 99999))])

        filename = prefix + timestamp
        return find_filename(self.tempdir,
                             filename)

    def _get_adapter_options(self):
        try:
            adapter = ISendAsPDFOptionsMaker(self.context)
        except TypeError:
            # There's no product implementing the adapter
            # available.
            return {}, None

        return adapter.getOptions(), adapter.overrideAll()

    def get_extra_options(self):
        """ Fetches the options defined in the request, the tool
        or an adapter.
        """
        # Options change depending on the pdf generator..
        try:
            transform_module = getattr(transforms, self.pdf_generator)
        except AttributeError:
            return []

        options = []
        tool_options = self.pdf_tool.make_options()
        adapter_options, adapter_overrides = self._get_adapter_options()

        opts_order = [self.request, tool_options]
        if adapter_overrides:
            opts_order.insert(0, adapter_options)
        else:
            opts_order.append(adapter_options)

        # First we check the options for which no value is
        # needed.
        # For each one, it is possible to define a --no-xxx
        # option.
        for opt_name in transform_module.simple_options:
            for opts in opts_order:
                if opts.get('--no-%s' % opt_name):
                    break

                # Check for None or False.  A zero may be a perfectly
                # valid option value that we want to add.
                if opts.get(opt_name, None) not in (None, False):
                    options.append('--%s' % opt_name)
                    break
        # Then we check values that expect a value.
        for opt_name in transform_module.valued_options:
            for opts in opts_order:
                opt_val = opts.get(opt_name, None)

                if opt_val is None:
                    continue

                # Value is put before the option name as we
                # insert them after in another list using l.insert(2, opt)
                if isinstance(opt_val, list):
                    for x in reversed(opt_val):
                        options.append(str(x))
                else:
                    options.append(str(opt_val))

                options.append('--%s' % opt_name)
                break

        return options

    def generate_pdf_file(self, source):
        """ Generates a PDF file from the given source
        (string containing the HTML source of a page).
        """
        filename = self.generate_temp_filename()
        if not filename:
            self.errors.append('filename_generation_failed')
            return

        try:
            transform_module = getattr(transforms, self.pdf_generator)
        except AttributeError:
            self.errors.append('wrong_generator_configuration')
            return

        self.filename = filename
        url = self.context.absolute_url()

        print_css = (self.pdf_tool.always_print_css or
                     self.context.portal_type in self.pdf_tool.print_css_types)

        # When the source is sent through Ajax, it's already encoded
        # as a utf-8 string.  When using it without javascript, the
        # source comes from a view, which always returns unicode.  In
        # that case we need to encode it.
        if isinstance(source, unicode):
            source = source.encode('utf-8')
        # Should we rewrite https links?  This option was added later,
        # so we check if the field is there.
        if getattr(self.pdf_tool, 'rewrite_https_links', False):
            source = source.replace('https://', 'http://')
        export_file, err = transform_module.html_to_pdf(
            source,
            self.tempdir,
            filename,
            url,
            print_css,
            self.get_extra_options())

        if err:
            self.errors.append('pdf_creation_failed')
            return

        self.pdf_tool.registerPDF(filename)
        self.pdf_file = export_file
        self.pdf_file.close()

    def make_pdf(self):
        """ Fetches the page source and generates a PDF.
        """
        source = self.get_page_source()
        if not source:
            self.errors.append('no_source')
        if not self.errors:
            self.generate_pdf_file(source)

    def show_error_message(self, error_name):
        """ Tells if an error message should be shown in the template.
        """
        return error_name in self.errors

    def class_for_field(self, fieldname):
        """ Returns the class that should be applied to a field
        in the forms displayed by the product.
        """
        base_class = 'field'
        error_class = ' error'
        if fieldname not in self.error_mapping:
            if fieldname in self.errors:
                base_class += error_class
            return base_class

        for error_name in self.error_mapping[fieldname]:
            if self.show_error_message(error_name):
                return base_class + error_class
        return base_class

    def check_pdf_accessibility(self):
        """ Check that the filename given in the request
        can be accessed by the user.
        """
        if 'pdf_name' not in self.request.form:
            # Should not happen.
            self.errors.append('file_not_specified')
            return

        filename = self.request.form['pdf_name']
        prefix = self.generate_filename_prefix()
        if not filename.startswith(prefix):
            # The user should not be able to see this file.
            self.errors.append('file_unauthorized')
            return

        if filename not in os.listdir(self.tempdir):
            self.errors.append('file_not_found')
            self.request.response.setStatus(404)
            return



from Products.Five.browser.pagetemplatefile import ZopeTwoPageTemplateFile

from jquery.pyproxy.plone import jquery, JQueryProxy

from collective.sendaspdf.browser.send import SendForm
from collective.sendaspdf.utils import update_relative_url


class SendAsPDFAjax(SendForm):
    """ This class contains a set of methods that are called
    with jquery.pyproxy.
    """

    def get_page_source(self):
        """ We override the get_page_source has this is
        sent by the ajax request.
        """
        return update_relative_url(self.request.form.get('page', ''),
                                   self.context)

    def add_popup(self, jq):
        jq.extend_grammar({'send_as_pdf_lightbox': []})

        jq('#send_as_pdf_popup').remove()
        jq('body').append('<div id="send_as_pdf_popup"></div>')

        # the self.index(self) looks really weird, but it does
        # not work in plone 4 with just self.index().
        jq('#send_as_pdf_popup').html(self.index(self))
        jq('#send_as_pdf_popup').send_as_pdf_lightbox()

        return jq

    def _show_send_form(self):
        jq = JQueryProxy()

        if 'page' not in self.request.form:
            # This should not happen.
            return jq

        self.make_pdf()
        if self.errors:
            self.index = ZopeTwoPageTemplateFile('templates/ajax.pt')
        else:
            self.index = ZopeTwoPageTemplateFile('templates/send_form.pt')

        return self.add_popup(jq)

    @jquery
    def show_send_form(self):
        # It's just to be able to use @@reload without being bothered
        # by the decorator.
        return self._show_send_form()

    def _send_mail(self):
        jq = JQueryProxy()
        self.check_form()

        if not self.errors:
            self.process_form()
            self.index = ZopeTwoPageTemplateFile('templates/ajax.pt')
            jq('#send_as_pdf_popup').html(self.index(self))
        else:
            # First update the fields class.
            for field in ['name_recipient',
                          'email_recipient',
                          'title',
                          'text',
                          'name',
                          'email']:

                if 'error' in self.class_for_field(field):
                    jq('#field_' + field).addClass('error')
                else:
                    jq('#field_' + field).removeClass('error')

            # We hide the previous errors.
            jq('.error_msg').addClass('dont-show')
            jq('.error_msg').removeClass('error_msg')

            for error in self.errors:
                jq('#error_' + error).removeClass('dont-show')
                jq('#error_' + error).addClass('error_msg')

        return jq

    @jquery
    def send_mail(self):
        return self._send_mail()

    def _download(self):
        jq = JQueryProxy()
        if 'page' not in self.request.form:
            # This should not happen.
            return jq

        jq.extend_grammar({'redirect': [[str, unicode]]})

        self.make_pdf()
        if self.errors:
            self.index = ZopeTwoPageTemplateFile('templates/ajax.pt')
            return self.add_popup(jq)

        jq('').redirect('%s/send_as_pdf_download?pdf_name=%s' % (
            self.context.absolute_url(),
            self.filename))
        return jq

    @jquery
    def download(self):
        return self._download()



""" Uninstall Profile
"""
from Products.CMFCore.utils import getToolByName


def uninstall(portal, reinstall=False):
    """ Uninstall profile setup
    """
    if not reinstall:
        setup_tool = getToolByName(portal, 'portal_setup')

        setup_tool.runAllImportStepsFromProfile(
            'profile-collective.sendaspdf:uninstall')

        portal.manage_delObjects(['portal_sendaspdf'])
        return "Ran all uninstall steps."



import re
import os
import base64

from zExceptions import Unauthorized
from Acquisition import aq_inner, aq_parent, aq_chain

from zope.component import getMultiAdapter

from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView

from Products.Archetypes.config import RENAME_AFTER_CREATION_ATTEMPTS

try:
    # Python 2.6 (maybe 2.5 too)
    import hashlib
    hashlib  # pyflakes
except ImportError:
    # Python 2.4
    import md5 as hashlib


def md5_hash(string, salt=''):
    return hashlib.md5(salt + string).hexdigest()


def decode_parameter(p):
    """ Decode a parameter/value from a URL format to
    a more readable version.
    Based on table seen in this page.
    http://www.blooberry.com/indexdot/html/topics/urlencoding.htm

    >>> from collective.sendaspdf.utils import decode_parameter
    >>> decode_parameter('kikoo%20lol')
    'kikoo lol'

    Well, if there is nothing to do, we do nothing.
    >>> decode_parameter('kikoolol')
    'kikoolol'

    It might mainly be used to decode the 'came_from' parameter.
    >>> url = 'http%3A//www.prettigpersoneel.nl/nerull-ii/%3Fview%3Demployees'
    >>> decode_parameter(url)
    'http://www.prettigpersoneel.nl/nerull-ii/?view=employees'

    We decode the '%' caracter last so we can not have confusion.
    >>> decode_parameter('%252C')
    '%2C'

    If we did not decode it last, we could have obtained ','.
    """
    table = {'24': '$',
             '26': '&',
             '2C': ',',
             '2F': '/',
             '3A': ':',
             '3B': ';',
             '3D': '=',
             '3F': '?',
             '40': '@',
             '20': ' ',
             '22': '"',
             '3C': '<',
             '3E': '>',
             '23': '#',
             '7B': '{',
             '7D': '}',
             '7C': '|',
             '5C': '\\',
             '5E': '^',
             '7E': '~',
             '5B': '[',
             '5D': ']',
             '60': '`'}

    for k in table:
        p = p.replace('%' + k, table[k])
    return p.replace('%25', '%')


def extract_from_url(url, context_url):
    """ Extracts the view name and the list of GET
    parameters from an URL, using the base context URL
    to extract the view name.

    >>> from collective.sendaspdf.utils import extract_from_url

    If the URL does not start with the context_url, wwe can
    not do anything.
    >>> extract_from_url('http://bla.com', 'https://bla.com')
    (None, None)

    If the two addresses are the same, we do not have any
    GET parameter or view.
    >>> context_url = 'http://bla.com/my_folder/my_context'
    >>> extract_from_url(context_url, context_url)
    ('', {})

    We can recognize the views with or without '@@' in front.
    >>> extract_from_url(context_url + '/my_view', context_url)
    ('my_view', {})

    >>> extract_from_url(context_url + '/@@my_view', context_url)
    ('my_view', {})

    It can also extract the list of GET parameters.
    >>> extract_from_url(context_url + '?p1=12', context_url)
    ('', {'p1': '12'})

    Even for multiple parameters of course.
    >>> extract_from_url(context_url + '?p1=12&p2=blabla', context_url)
    ('', {'p2': 'blabla', 'p1': '12'})

    If a GET parameter is present twice, then the corresponding value
    in the dictionary will be a list containing all values.
    >>> extract_from_url(context_url + '?p1=12&p2=blabla&p2=blublu',
    ...                  context_url)
    ('', {'p2': ['blabla', 'blublu'], 'p1': '12'})

    Check issue 11, where there are problems with colons in the
    querystring.  Could be better to use urlparse.parse_qs
    See https://github.com/zestsoftware/collective.sendaspdf/issues/11
    >>> extract_from_url(context_url + '?ref=562a&to_date=2013-12-31 00:00',
    ...                  context_url)
    ('', {'ref': '562a', 'to_date': '2013-12-31 00:00'})
    >>> extract_from_url(context_url + '?ref=562a&to_date=2013-12-31',
    ...                  context_url)
    ('', {'ref': '562a', 'to_date': '2013-12-31'})

    """
    if not url.startswith(context_url):
        return None, None

    carac_class = '[A-Za-z0-9_ %/\\.\\-:+]'

    reg = r'^' + context_url + '/? ' + \
          '?((@@)?(' + carac_class + '*)/?)?' + \
          '(\\?((' + carac_class + '+=' + carac_class + '*&?)*))?' + \
          '(#.*)?$'
    match = re.split(reg, url)
    if len(match) == 1:
        return None, None

    view_name = match[3]

    # We still need to extract parameters more properly.
    get_params = {}
    if match[6]:
        for couple in match[5].split('&'):
            try:
                key, value = couple.split('=')
                key = decode_parameter(key)
                value = decode_parameter(value)

                if key in get_params:
                    if not type(get_params[key]) == list:
                        get_params[key] = [get_params[key]]
                    get_params[key].append(value)
                else:
                    get_params[key] = value
            except:
                pass
    return view_name, get_params


def find_filename(path, filename, extension='pdf'):
    """ Finds a non-conflicting filename in the
    directory given by path.

    >>> import os
    >>> from collective.sendaspdf.utils import find_filename

    If the path is not correct, returns nothing.
    If this test fails, I'm not responsible for the fact that you
    have a 'kikoolol' directory in the root of your system ;)
    >>> find_filename('/kikoolol', '')

    Some faking for the tests.
    >>> def fake_listdir(path):
    ...     return ['file.pdf', 'file1.pdf', 'file2.pdf', 'file3.pdf',
    ...             'file.html', 'file1.html', 'file2.html']

    >>> os.old_listdir = os.listdir
    >>> os.listdir = fake_listdir

    A example without conflict.
    >>> find_filename('', 'my_file')
    'my_file.pdf'

    No conflict and a custom extension.
    >>> find_filename('', 'my_file', 'html')
    'my_file.html'

    If there is a conflict, the function adds a number
    at the end.
    >>> find_filename('', 'file')
    'file4.pdf'

    >>> find_filename('', 'file', 'html')
    'file3.html'

    """
    try:
        files = os.listdir(path)
    except:
        # This shall not happen, except if the path has not
        # been set correctly.
        return

    # We check the file name is not already used in the directory.
    # If so, we try to prepend a number at the end.
    if filename + '.' + extension in files:
        postfix = 1
        while postfix <= RENAME_AFTER_CREATION_ATTEMPTS:
            if '%s%s.%s' % (filename, postfix, extension) in files:
                postfix += 1
                continue

            filename = '%s%s' % (filename, postfix)
            break

        if postfix > RENAME_AFTER_CREATION_ATTEMPTS:
            return

    return filename + '.' + extension

img_sizes = {'image': [],
             'image_listing': [16, 16],
             'image_icon': [32, 32],
             'image_tile': [64, 64],
             'image_thumb': [128, 128],
             'image_mini': [200, 200],
             'image_preview': [400, 400],
             'image_large': [768, 768]}


def get_object_from_url(context, path):
    """ Returns a tuple object, view name, image size, unparsed items.

    View tests/utils.txt for samples.
    """
    obj = context

    for position, element in enumerate(path):
        if element == '..':
            parent = aq_parent(aq_inner(obj))
            if parent is not None:
                obj = parent
            continue

        try:
            obj = getattr(obj, element)
        except AttributeError:
            # Three possibilities here:
            # - the path is broken
            # - the element is a view name
            # - the element defines the image size (in case of images)

            if element in img_sizes:
                return obj, None, element, path[position + 1:]

            try:
                # We sometimes have the problem with Plone 3 where
                # Acquisition seems to not have an effect on getattr
                # or something like that ...
                obj = getattr(aq_parent(obj), element)
                continue
            except AttributeError:
                pass

            # To test the views, we'll use the full aq_chain.
            for ancestor in aq_chain(aq_inner(obj)):
                try:
                    view = ancestor.restrictedTraverse(str(element))
                    if isinstance(view, BrowserView):
                        return ancestor, element, None, path[position + 1:]

                    # Ho, Plone 3 again. An object in the skin folder can not
                    # be accessed via getattr (when in a folder, works when
                    # in a document ...).
                    obj = view
                    if position == len(path) - 1:
                        return obj, None, None, None
                    continue

                except (Unauthorized, AttributeError, KeyError, ):
                    pass

            # Ok, we really can't find it now.
            return None, None, None, None

    return obj, None, None, None


def update_relative_url(source, context, embedded_images=True):
    relative_exp = re.compile('((href|src)="([a-zA-Z0-9_=&\-\.\/@\?]+)")',
                              re.MULTILINE | re.I | re.U)
    protocol_exp = re.compile('^(\w+:\/\/).*$')
    image_exp = re.compile('^.*\.(jpg|jpeg|gif|png).*$')
    anchor_exp = re.compile('((href)="(#[^"]+)")',
                            re.MULTILINE | re.I | re.U)

    items = relative_exp.findall(source)
    original_url = context.absolute_url()

    mtool = getToolByName(context, 'portal_membership')

    anchor_items = anchor_exp.findall(source)
    for anchor_item in anchor_items:
        attr = anchor_item[1]
        value = anchor_item[2]
        source = source.replace('href="%s"' % value,
                                'href="%s%s"' % (original_url, value))

    for item in items:
        attr = item[1]
        value = item[2]

        if protocol_exp.match(value):
            # That should not happen as ':' should not be recognized
            # by relative_exp.
            continue

        if '?' in value:
            # Just to make things funnier, we have get parameters...
            split = value.split('?')
            value = split[0]
            get_params = '?'.join(split[1:])
        else:
            get_params = None

        path = value.split('/')

        portal = getMultiAdapter((context, context.REQUEST),
                                 name="plone_portal_state").portal()
        portal_path = portal.absolute_url()
        default_replacement = '%s=%s/%s' % (attr, portal_path, value)
        if get_params:
            default_replacement = '%s?%s' % (default_replacement, get_params)

        linked_obj, view, img_size, left_path = get_object_from_url(context,
                                                                    path)

        if linked_obj is None:
            source = source.replace('%s="%s"' % (attr, value),
                                    default_replacement)
            continue

        replacement = linked_obj.absolute_url()
        if view:
            replacement += '/%s' % view

        is_img = image_exp.match(value)
        is_src = (attr == 'src')
        no_context = (linked_obj != context)
        can_view = mtool.checkPermission('View', linked_obj)

        if is_img and is_src and no_context and can_view:
            if not embedded_images:
                replacement = '%s' % linked_obj.absolute_url()
                if view in ['images', '@@images']:
                    # Image integrated with Plone 4/TinyMCE.
                    replacement += '/@@images/' + '/'.join(left_path)
                elif img_size:
                    replacement += '/' + img_size
            else:
                try:
                    filename = linked_obj.getImage().getFilename()
                    filetype = filename.split('.')[-1]
                    content = linked_obj.getImageAsFile().read()

                    # That's the default image (in it's full size). We'll now
                    # try to resize it.
                    is_img_view = view in ['images', '@@images']
                    is_img_size = (left_path and
                                   'image_%s' % left_path[-1] in img_sizes)

                    if is_img_view and is_img_size:
                        # Plone 4
                        img_view = linked_obj.restrictedTraverse('@@images')
                        img_w, img_h = img_sizes['image_%s' % left_path[-1]]
                        content = img_view.scale(height=img_h,
                                                 width=img_w).data

                    elif img_size:
                        resized = linked_obj.restrictedTraverse(str(img_size))
                        content = resized.data

                except AttributeError:
                    # that's an image in the skin folder.
                    try:
                        filetype = linked_obj.filename.split('.')[-1]
                    except AttributeError:
                        filetype = linked_obj._filepath.split('.')[-1]
                    content = linked_obj._readFile(False)

                replacement = 'data:image/%s;base64,%s' % (
                    filetype,
                    base64.encodestring(content))

        if get_params:
            replacement += '?%s' % (get_params)

        replacement = '%s="%s"' % (attr, replacement)

        source = source.replace('%s="%s"' % (attr, item[2]),
                                replacement)

    return source



from collective.sendaspdf.browser.base import BaseView


class RealURLView(BaseView):
    """ We need this view to build the 'send as pdf'
    action menu.
    """
    def __call__(self):
        base = self.context.REQUEST['ACTUAL_URL']
        method = self.context.REQUEST['REQUEST_METHOD']

        get_params = ''
        if method == 'GET':
            get_params = '&'.join(
                ['%s=%s' % (k, v) for k, v in self.context.REQUEST.form.items()
                 if k != '-C' and not (k == 'test' and v == '')])

        if get_params:
            base += '?' + get_params

        return base



from setuptools import setup, find_packages

setup(name='collective.sendaspdf',
      version='3.0',
      description="An open source product for Plone to download or email a " +
                  "page seen by the user as a PDF file.",
      long_description=(open("README.rst").read() + "\n" +
                        open("CHANGES.rst").read()),
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Framework :: Plone",
          "Framework :: Plone :: 3.3",
          "Framework :: Plone :: 4.0",
          "Framework :: Plone :: 4.1",
          "Framework :: Plone :: 4.2",
          "Framework :: Plone :: 4.3",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.4",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          ],
      keywords='html pdf converter',
      author='Zest Software',
      author_email='v.pretre@zestsoftware.nl',
      url='http://github.com/zestsoftware/collective.sendaspdf',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'jquery.pyproxy>=0.3',
      ],
      extras_require={
          'test': ['pdfminer',
                   'Products.PloneTestCase',
                   'zope.app.testing',
                   'zope.securitypolicy',
                   ],
          },
      entry_points="""
      # -*- Entry points: -*-

      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



import logging
from Products.CMFCore.utils import getToolByName

logger = logging.getLogger('collective.sendaspdf')


def update_control_panel_and_tool(context):
    """Update our control panel action and tool.

    It may need a different icon expression (gif versus png).

    We first added it in the Plone category, but it should be in the
    Products category.

    For the tool: set the content icon properly, again gif versus png.

    For Plone 3, we do not touch the icons.  You might have a wrong
    icon then, but only when you install sendaspdf for the first time,
    which is more likely to happen on Plone 4.
    """
    controlPanel = getToolByName(context, 'portal_controlpanel')
    action = controlPanel.getActionObject('Plone/sendaspdf')
    if action is not None:
        action.category = 'Products'
        logger.info('Set sendaspdf controlpanel category to Products.')
    else:
        action = controlPanel.getActionObject('Products/sendaspdf')

    try:
        # Plone 4 or higher
        import plone.app.upgrade
        plone.app.upgrade  # pyflakes
        icon = 'string:${portal_url}/pdf_icon.png'
    except ImportError:
        # Plone 3
        # icon = 'string:${portal_url}/pdf_icon.gif'
        # Actually, we will do nothing anymore.
        return

    if action.icon_expr.text != icon:
        action.icon_expr.text = icon
        logger.info('Set sendaspdf controlpanel icon expression to %r', icon)

    # Now the tool.
    types = getToolByName(context, 'portal_types')
    typeinfo = types.getTypeInfo('SendAsPDFTool')
    if typeinfo is None:
        return
    if hasattr(typeinfo, 'icon_expr'):
        # Plone 4
        if typeinfo.icon_expr_object.text != icon:
            typeinfo.icon_expr_object.text = icon
            logger.info('Set sendaspdf tool icon expression to %r', icon)



import doctest
from unittest import TestSuite
from Testing.ZopeTestCase.zopedoctest import ZopeDocFileSuite

from collective.sendaspdf.tests.base import SendAsPDFTestCase

OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE)


def test_suite():
    func = ZopeDocFileSuite('func.txt',
                            package='collective.sendaspdf.tests',
                            optionflags=OPTIONFLAGS,
                            test_class=SendAsPDFTestCase)
    adapter = ZopeDocFileSuite('adapter.txt',
                               package='collective.sendaspdf.tests',
                               optionflags=OPTIONFLAGS,
                               test_class=SendAsPDFTestCase)

    return TestSuite((func, adapter, ))



from zope.component import adapts

from Products.GenericSetup.interfaces import ISetupEnviron
from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects
from Products.GenericSetup.utils import XMLAdapterBase

from Products.CMFCore.utils import getToolByName
from collective.sendaspdf.tool import ISendAsPDFTool


class SendAsPdfToolXMLAdapter(XMLAdapterBase):
    adapts(ISendAsPDFTool, ISetupEnviron)

    _LOGGER_ID = 'sendaspdf'

    name = 'sendaspdftool'

    exportfield = {
        'pdf_generator': (lambda x: str(x),
                          lambda x: x),
        'tempdir': (lambda x: str(x),
                    lambda x: x),
        'excluded_browser_attachment': (lambda x: ';'.join(x),
                                        lambda x: tuple(x.split(';'))),
        'salt': (lambda x: str(x),
                 lambda x: x),
        'mail_title': (lambda x: str(x),
                       lambda x: x),
        'mail_content': (lambda x: str(x),
                         lambda x: x),
        'filename_in_mail': (lambda x: str(x),
                             lambda x: x),
        'always_print_css': (lambda x: str(x),
                             lambda x: bool(x)),
        'print_css_types': (lambda x: ';'.join(x),
                            lambda x: tuple(x.split(';')))}

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._getObjectNode('object')
        node.appendChild(self._extractPredicates())

        self._logger.info('Send as pdf tool exported.')
        return node

    def _importNode(self, node):
        """Import the object from the DOM node.
        """
        if self.environ.shouldPurge():
            self._purgePredicates()

        self._initPredicates(node)

        self._logger.info('Send as pdf tool imported.')

    def _extractPredicates(self):
        fragment = self._doc.createDocumentFragment()

        for k, functions in self.exportfield.iteritems():
            serialize = functions[0]
            v = serialize(self.context[k])
            child = self._doc.createElement('property')
            child.setAttribute('name', k)
            child.setAttribute('value', v)
            fragment.appendChild(child)

        return fragment

    def _purgePredicates(self):
        self.context.__init__()

    def _initPredicates(self, node):
        out = {}
        for child in node.childNodes:
            if child.nodeName != 'property':
                continue

            name = child.getAttribute('name')
            if name in self.exportfield:
                deserialize = self.exportfield[name][1]
                value = deserialize(child.getAttribute('value'))
                out[str(name)] = value

        self.context.update(**out)


def importProviders(context):
    """Import actions tool.
    """
    site = context.getSite()
    tool = getToolByName(site, 'portal_sendaspdf', None)
    if tool is None:
        logger = context.getLogger('sendaspdf')
        logger.debug('Nothing to import.')
        return

    importObjects(tool, '', context)


def exportProviders(context):
    """Export actions tool.
    """
    site = context.getSite()
    tool = getToolByName(site, 'portal_sendaspdf', None)
    if tool is None:
        logger = context.getLogger('sendaspdf')
        logger.debug('Nothing to export.')
        return
    exportObjects(tool, '', context)



import unittest
import doctest

from Testing.ZopeTestCase.zopedoctest import ZopeDocFileSuite

from collective.sendaspdf.tests.base import OPTIONFLAGS
from collective.sendaspdf.tests.base import SendAsPDFTestCase


def test_suite():
    utils_py = doctest.DocFileSuite('utils.py',
                                    optionflags=OPTIONFLAGS,
                                    package='collective.sendaspdf')
    utils_txt = ZopeDocFileSuite('utils.txt',
                                 package='collective.sendaspdf.tests',
                                 optionflags=OPTIONFLAGS,
                                 test_class=SendAsPDFTestCase)
    return unittest.TestSuite([utils_py, utils_txt])



# Defines the html_to_pdf method using WKHTMLTOPDF:
# http://code.google.com/p/wkhtmltopdf/
import os
import signal
import subprocess
import logging

from tempfile import TemporaryFile
from threading import Timer

from collective.sendaspdf.utils import find_filename

logger = logging.getLogger('collective.sendaspdf')

wk_command = os.environ.get('WKHTMLTOPDF_PATH')
if wk_command:
    logger.info('wkhtmltopdf found at  %s: ' % wk_command)
else:
    wk_command = 'wkhtmltopdf'
    logger.warn("wkhtmltopdf path unknown, hope it's in the path")

simple_options = ['book', 'collate',
                  'disable-external-links', 'disable-internal-links',
                  'disable-pdf-compression', 'disable-smart-shrinking',
                  'forms', 'grayscale', 'lowquality', 'no-background',
                  'header-line', 'footer-line',
                  'toc', 'toc-disable-back-links', 'toc-disable-links']

valued_options = ['copies', 'cover', 'dpi',
                  'margin-top', 'margin-bottom', 'margin-left', 'margin-right',
                  'minimum-font-size', 'orientation',
                  'page-height', 'page-offset', 'page-size', 'page-width',
                  'header-font-name', 'header-html', 'header-font-size',
                  'header-spacing', 'header-left', 'header-center',
                  'header-right', 'footer-font-name', 'footer-html',
                  'footer-font-size', 'footer-spacing',
                  'footer-left', 'footer-center', 'footer-right',
                  'toc-depth', 'toc-header-text', 'cookie']


def html_to_pdf(source, export_dir, filename,
                original_url, use_print_css, extra_options=[]):
    # First we need to store the source in a temporary
    # file.
    html_filename = find_filename(export_dir,
                                  filename,
                                  'html')
    if not html_filename:
        return None, ['no_filename_temp_html']

    html_file = file('%s/%s' % (export_dir, html_filename),
                     'wb')

    html_file.write(source)
    html_file.close()

    # Run the wkhtmltopdf command.
    args = [wk_command,
            '--disable-javascript',
            '--encoding',
            'utf-8',
            'file://%s/%s' % (export_dir, html_filename),
            '%s/%s' % (export_dir, filename)]

    if use_print_css:
        args.insert(4, '--print-media-type')

    for opt in extra_options:
        args.insert(4, opt)

    try:
        proc = subprocess.Popen(args,
                                stdin=TemporaryFile(),
                                stdout=TemporaryFile(),
                                stderr=TemporaryFile())

        if hasattr(proc, 'kill'):
            # Plone 4
            timer = Timer(10, proc.kill)
        else:
            # Plone 3
            timer = Timer(10, lambda pid: os.kill(pid, signal.SIGKILL),
                          [proc.pid])
        timer.start()
        proc.communicate()
        timer.cancel()
    except Exception, err:
        logger.error('Running wkhtmltopdf failed. wkhtmltopdf cmd: %s. '
                     'Error: %s', args, err)
        return None, ['pdf_generation_failed']

    try:
        os.remove('%s/%s' % (export_dir, html_filename))
    except IOError, err:
        logger.error('Temp file does not exist: %s. wkhtmltopdf cmd: %s',
                     err, args)

    try:
        pdf_file = file('%s/%s' % (export_dir, filename), 'r')
    except IOError, err:
        logger.error('No PDF output file: %s. wkhtmltopdf cmd: %s', err, args)
        return None, ['pdf_generation_failed']

    return pdf_file, None
